import logging
from typing import Tuple

import trio

from .openrover_data import MOTOR_EFFORT_FORMAT, OPENROVER_DATA_ELEMENTS
from .serial_trio import SerialTrio
from .util import OpenRoverException

SERIAL_START_BYTE = bytes([253])


def encode_packet(*args: bytes):
    payload = b''.join(args)
    return SERIAL_START_BYTE + payload + bytes([checksum(payload)])


def checksum(values):
    return 255 - sum(values) % 255


class OpenRoverProtocol():
    def __init__(self, serial: SerialTrio):
        """Low-level communication for OpenRover"""
        self._serial = serial
        self._read_lock = trio.StrictFIFOLock()

    async def iter_packets(self):
        while True:
            try:
                msg = await self._read_one_packet()
                yield msg
            except OpenRoverException as e:
                logging.exception(str(e))

    async def _read_one_packet(self) -> Tuple[int, bytes]:
        async with self._read_lock:
            _ = await self._serial.read_until(SERIAL_START_BYTE)
            packet = SERIAL_START_BYTE + await self._serial.read_exactly(4)
            assert len(packet) == 5
            if packet[4] == checksum(packet[1:4]):
                k = packet[1]
                v = OPENROVER_DATA_ELEMENTS[k].data_format.unpack(packet[2:4])
                return (k, v)
            else:
                raise OpenRoverException('Bad checksum. Discarding data ' + repr(packet))

    async def write(self, motor_left, motor_right, flipper, arg1, arg2):
        binary = encode_packet(MOTOR_EFFORT_FORMAT.pack(motor_left),
                               MOTOR_EFFORT_FORMAT.pack(motor_right),
                               MOTOR_EFFORT_FORMAT.pack(flipper),
                               bytes([arg1]),
                               bytes([arg2]))
        await self._serial.write(binary)



import asyncio
from asyncio import StreamWriter, StreamReader
from enum import IntEnum
import logging
import struct
from typing import Any, Tuple

from dataclasses import dataclass

_START_OF_FRAME = 0xf7
_END_OF_FRAME = 0x7f
_ESC = 0xf6
_ESC_XOR = 0x20


class BootyProtocolError(RuntimeError):
    pass


class BootyCommand(IntEnum):
    READ_PLATFORM = 0x00
    READ_VERSION = 0x01
    READ_ROW_LEN = 0x02
    READ_PAGE_LEN = 0x03
    READ_PROG_LEN = 0x04
    READ_MAX_PROG_SIZE = 0x05
    READ_APP_START_ADDRESS = 0x06
    READ_BOOT_START_ADDRESS = 0x07
    READ_ADDR = 0x20
    READ_MAX = 0x21
    ERASE_PAGE = 0x10
    WRITE_ROW = 0x30
    WRITE_MAX = 0x31
    START_APP = 0x40


@dataclass
class BootyDeviceMetadata():
    platform: str
    version: str
    row_len: int
    page_len: int
    prog_len: int
    max_prog_size: int
    app_start_address: int
    boot_start_address: int


class BootyProtocol:
    _stream_reader: asyncio.StreamReader = None
    _process_packets_task = None

    def __init__(self, reader: StreamReader, writer: StreamWriter):
        self._queue = asyncio.Queue()
        self._reader = reader
        self._writer = writer

    # region Booty Protocol actions
    async def get_device_metadata(self):
        platform = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_PLATFORM))
        version = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_VERSION))
        row_len = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_ROW_LEN))
        page_len = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_PAGE_LEN))
        prog_len = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_PROG_LEN))
        max_prog_size = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_MAX_PROG_SIZE))
        app_start_address = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_APP_START_ADDRESS))
        boot_start_address = asyncio.ensure_future(self._cmd_request_data(BootyCommand.READ_BOOT_START_ADDRESS))
        await asyncio.wait([platform, version, row_len, page_len, prog_len, max_prog_size, app_start_address, boot_start_address], timeout=2)
        metadata = BootyDeviceMetadata(
            platform=platform.result(),
            version=version.result(),
            row_len=row_len.result(),
            page_len=page_len.result(),
            prog_len=prog_len.result(),
            max_prog_size=max_prog_size.result(),
            app_start_address=app_start_address.result(),
            boot_start_address=boot_start_address.result()
        )

        logging.info(f'Platform metadata = {metadata}')
        return metadata

    def cmd_start_app(self):
        self._cmd_write(BootyCommand.START_APP)

    def cmd_write_row(self, addr, instructions):
        self._cmd_write(BootyCommand.WRITE_ROW, addr, instructions)

    def cmd_write_max(self, addr, instructions):
        self._cmd_write(BootyCommand.WRITE_MAX, addr, instructions)

    def cmd_erase_page(self, addr):
        self._cmd_write(BootyCommand.ERASE_PAGE, addr)

    async def cmd_read_addr(self, addr):
        await self._cmd_write(BootyCommand.READ_ADDR, addr)
        return await self._cmd_expect_response(BootyCommand.READ_ADDR)

    async def cmd_read_max(self, addr):
        await self._cmd_write(BootyCommand.READ_MAX, addr)
        return await self._cmd_expect_response(BootyCommand.READ_MAX)

    # endregion

    def _pack_data(self, verb, *args):
        payload = struct.pack('!H')
        if verb == BootyCommand.ERASE_PAGE:
            payload += struct.pack('!H', *args)
        elif verb in [BootyCommand.WRITE_ROW, BootyCommand.WRITE_MAX]:
            addr, data = args
            payload += struct.pack('!H', addr)
            payload += b''.join(struct.pack('!L', instr) for instr in data)
        else:
            # most request packets have no associated data
            assert len(args) == 0

        return payload

    def _unpack_data(self, payload: bytes) -> Tuple[BootyCommand, Any]:
        verb = BootyCommand(*struct.unpack('!H', payload[:2]))
        data = payload[2:]

        if verb in [BootyCommand.READ_PLATFORM, BootyCommand.READ_VERSION]:
            # interpret data as a string
            return (verb, data.rstrip(b'\0').decode('utf-8'))
        elif verb in [BootyCommand.READ_ADDR, BootyCommand.READ_MAX]:
            # interpret data as a list of instructions (uint32's)
            return (verb, [x for x, in struct.iter_unpack('!L', data)])
        else:
            # interpret data as a uint32
            return (verb, *struct.unpack('!L', data))

    def _remove_esc_chars(self, raw_message: bytes):
        if _START_OF_FRAME in raw_message:
            raise BootyProtocolError('Raw message contains start of frame character', raw_message)
        if _END_OF_FRAME in raw_message:
            raise BootyProtocolError('Raw message contains start of frame character', raw_message)
        if raw_message[-1] == _ESC:
            raise BootyProtocolError('Raw message ends with escape character', raw_message)

        message = bytearray()
        escape_next = False
        for c in raw_message:
            if escape_next:
                message.append(c ^ _ESC_XOR)
                escape_next = False
            elif c == _ESC:
                escape_next = True
            else:
                message.append(c)
        return bytes(message)

    async def _cmd_request_data(self, verb: BootyCommand) -> asyncio.Future:
        assert verb in [BootyCommand.READ_PLATFORM, BootyCommand.READ_VERSION, BootyCommand.READ_ROW_LEN, BootyCommand.READ_PAGE_LEN, BootyCommand.READ_PROG_LEN, BootyCommand.READ_MAX_PROG_SIZE,
                        BootyCommand.READ_APP_START_ADDRESS, BootyCommand.READ_BOOT_START_ADDRESS]
        await self._cmd_write(verb)

        reply_verb, data = await self._queue.get()
        if verb != reply_verb:
            raise BootyProtocolError(f'Requested data type {verb} but received {reply_verb} in response', data)
        return data

    async def _cmd_expect_response(self, expect_verb):
        reply_verb, data = await self._queue.get()
        if expect_verb != reply_verb:
            raise BootyProtocolError(f'Requested data type {expect_verb} but received {reply_verb} in response', data)
        return data

    async def _cmd_write(self, verb, *args):
        self._writer.write(
            bytes([_START_OF_FRAME, *self._make_tx_packet(verb, *args), _END_OF_FRAME])
        )
        await self._writer.drain()

    def _make_tx_packet(self, verb, *args):
        """Create a data packet, not including framing bytes"""
        pk = bytes([verb] + self._pack_data(verb, *args))
        pk = pk + self._fletcher16_checksum(pk)
        pk = pk.replace(bytes([_ESC]), bytes([_ESC, _ESC_XOR ^ _ESC]))
        pk = pk.replace(bytes([_START_OF_FRAME]), bytes([_ESC, _ESC_XOR ^ _START_OF_FRAME]))
        pk = pk.replace(bytes([_END_OF_FRAME]), bytes([_ESC, _ESC_XOR ^ _END_OF_FRAME]))
        return pk

    def _fletcher16_checksum(self, data: bytes):
        sum1 = 0
        sum2 = 0
        for b in data:
            sum1 = (sum1 + b) % 256
            sum2 = (sum2 + sum1) % 256
        return bytes([sum1, sum2])

    async def _receive_packet(self):
        before_frame_contents = await self._stream_reader.readuntil(bytes([_START_OF_FRAME]))[:-1]
        if len(before_frame_contents):
            logging.warning(f'Discarding data before start of frame {before_frame_contents}')
        raw_message = await self._stream_reader.readuntil(bytes([_END_OF_FRAME]))[:-1]
        logging.debug(f'Processing packet {raw_message}')

        message = self._remove_esc_chars(raw_message)
        if len(message) < 3:
            raise BootyProtocolError('Packet too short', raw_message)
        payload = message[:-2]
        checksum = message[-2:]
        expected_checksum = self._fletcher16_checksum(payload)
        if checksum != expected_checksum:
            raise BootyProtocolError('Packet failed checksum check', raw_message)

        return self._unpack_data(payload)

    async def iter_messages(self):
        while True:
            try:
                msg = await self._receive_packet()
                yield msg
            except BootyProtocolError as e:
                logging.warning(f'Skipping packet because of error {e}')



from pathlib import Path
import shutil
from subprocess import list2cmdline

import pytest
import trio

from .openrover_protocol import OpenRoverProtocol
from .serial_trio import open_first_possible_rover_device

device = pytest.fixture(open_first_possible_rover_device)


@pytest.fixture
def powerboard_firmware_file():
    p = Path('test-resources/PowerBoard-1.5.0.hex')
    assert p.is_file()
    return p


@pytest.fixture
def booty_exe():
    p = Path(shutil.which('booty'))
    assert p.is_file()
    return p


async def test_reboot(device):
    o = OpenRoverProtocol(device)
    assert isinstance(o, OpenRoverProtocol)
    # set a long timeout in case rover is already in bootloader

    try:
        with trio.fail_after(2):
            await o.write(0, 0, 0, 10, 40)
            i, version = await o.iter_packets().__anext__()
            assert i == 40
    except trio.TooSlowError:
        pytest.fail('Rover did not respond. does it have firmware?')

    try:
        await o.write(0, 0, 0, 230, 1)  # reboot the device

        await o.write(0, 0, 0, 10, 40)
        with pytest.raises(trio.TooSlowError):
            with trio.fail_after(2):
                await o.iter_packets().__anext__()

    finally:
        # need to wait for device to come back up
        await trio.sleep(10)


async def test_bootloader(powerboard_firmware_file, booty_exe):
    # note this test can take a loooong time

    with trio.fail_after(1):
        async with open_first_possible_rover_device() as device:
            o = OpenRoverProtocol(device)
            await o.write(0, 0, 0, 230, 0)
            port = device.port

    # flash rover firmware
    args = [
        str(booty_exe),
        '--port', str(port),
        '--baudrate', '57600',
        '--hexfile', str(powerboard_firmware_file.absolute()),
        '--erase',
        '--load',
        '--verify'
    ]

    print('running bootloader: ' + list2cmdline(args))

    with trio.fail_after(60 * 15):

        async with trio.subprocess.Process(args, stdout=trio.subprocess.PIPE, stderr=trio.subprocess.PIPE) as booty:
            async with trio.open_nursery() as nursery:
                async def check_stdout():
                    line_generator = stream_to_lines(booty.stdout)
                    lines = []
                    try:
                        while True:
                            with trio.fail_after(15):
                                a_line = await line_generator.__anext__()
                            print('got line: ' + a_line)
                            lines.append(a_line)
                    except trio.TooSlowError:
                        pytest.fail(f'booty became unresponsive after output {lines}')
                    except StopAsyncIteration:
                        pass
                    await trio.sleep(1)
                    assert any(['device verified' in a_line for a_line in lines])

                async def check_stderr():
                    error_output = await stream_to_string(booty.stderr)
                    if 'device not responding' in error_output:
                        pytest.fail('Rover did not respond to booty. Does it have a bootloader?')
                    assert error_output.strip() == ''

                async def check_retcode():
                    assert await booty.wait() == 0

                nursery.start_soon(check_stderr)
                nursery.start_soon(check_stdout)
                nursery.start_soon(check_retcode)

    async with open_first_possible_rover_device() as device:
        await trio.sleep(15)
        o = OpenRoverProtocol(device)
        await o.write(0, 0, 0, 10, 40)
        with pytest.raises(trio.TooSlowError):
            with trio.fail_after(2):
                k, version = await o.iter_packets().__anext__()
        assert k == 40
        assert (version.major, version.minor, version.patch) == (1, 5, 0)


async def stream_to_string(stream: trio.abc.ReceiveStream):
    buf = ''
    async with stream:
        while True:
            new_bytes = await stream.receive_some(1)
            if not new_bytes:
                break
            buf += new_bytes.decode()
    return buf


async def stream_to_lines(stream: trio.abc.ReceiveStream):
    buf = ''
    async with stream:
        while True:
            new_bytes = await stream.receive_some(1)
            if not new_bytes:
                break
            buf += new_bytes.decode()
            *some_lines, buf = buf.splitlines()
            for line in some_lines:
                yield line
        if buf != '':
            yield buf



import trio

from .openrover_data import OpenRoverFirmwareVersion
from .openrover_protocol import OpenRoverProtocol
from .serial_trio import open_first_possible_rover_device

n = 800


async def test_protocol_write_read_immediate():
    n_received = 0

    async with open_first_possible_rover_device() as port:
        protocol = OpenRoverProtocol(port)
        messages = protocol.iter_packets()
        for i in range(n):
            with trio.fail_after(1):
                await protocol.write(0, 0, 0, 10, 40)
                key, version = await messages.__anext__()
                assert key == 40
                assert isinstance(version, OpenRoverFirmwareVersion)
                assert isinstance(version.value, int)
                assert 0 < version.value
                n_received += 1

    print(f'success ratio {n_received / n}')
    assert 0.9 < n_received / n <= 1


async def test_protocol_writes_then_reads():
    n_received = 0

    async with open_first_possible_rover_device() as port:
        protocol = OpenRoverProtocol(port)
        messages = protocol.iter_packets()

        async with trio.open_nursery() as nursery:
            for i in range(n):
                nursery.start_soon(protocol.write, 0, 0, 0, 10, 40)
            try:
                for i in range(n):
                    with trio.fail_after(1):
                        key, version = await messages.__anext__()
                        assert key == 40
                        assert isinstance(version, OpenRoverFirmwareVersion)
                        assert isinstance(version.value, int)
                        assert 0 < version.value
                        n_received += 1
            except trio.TooSlowError:
                pass

    print(f'success ratio {n_received / n}')
    assert 0.9 < n_received / n <= 1



from math import isclose
import statistics
import time

import pytest
import trio

from .openrover_data import OpenRoverFirmwareVersion
from .rover import OpenRover
from .util import OpenRoverException


@pytest.fixture
async def rover():
    async with OpenRover() as o:
        yield o


async def test_find_openrover(rover):
    async with OpenRover() as o:
        assert o.port is not None


def test_create():
    o = OpenRover()
    assert o is not None


async def test_get_version(rover):
    assert await rover.get_data(40) is not None


async def test_recover_from_bad_data(rover):
    rover._rover_protocol._writer.write(b'test' * 20)

    for i in range(3):
        try:
            result = await rover.get_data(40)
            if result is not None:
                return
        except Exception as e:
            pass
    assert False


async def test_missing_device():
    rover = OpenRover(port='missingdevice')
    with pytest.raises(OpenRoverException):
        try:
            await rover.aopen()
        except Exception as e:
            raise


async def test_build_number():
    async with OpenRover() as rover:
        build_no = await rover.get_data(40)
        assert build_no is not None
        assert isinstance(build_no, OpenRoverFirmwareVersion)
        assert 0 <= build_no.major < 100
        assert 0 <= build_no.minor < 100
        assert 0 <= build_no.patch < 100


async def test_encoder_counts():
    async with OpenRover() as rover:
        enc_counts_1 = (await rover.get_data(14), await rover.get_data(16))
        await trio.sleep(0.3)
        enc_counts_2 = (await rover.get_data(14), await rover.get_data(16))
        assert enc_counts_1 == enc_counts_2

        rover.set_motor_speeds(0.2, 0.2, 0.2)
        rover.send_speed()
        await trio.sleep(0.3)

        enc_counts_3 = (await rover.get_data(14), await rover.get_data(16))
        enc_diff = ((enc_counts_3[0] - enc_counts_2[0]) % (2 ** 16),
                    (enc_counts_3[1] - enc_counts_2[1]) % 2 ** 16)
        assert 0 < enc_diff[0] < 200
        assert 0 < enc_diff[1] < 200


async def test_encoder_intervals_still(rover):
    await trio.sleep(2)
    enc_counts_left = []
    enc_counts_right = []
    enc_intervals_left = []
    enc_intervals_right = []
    rover.set_motor_speeds(0, 0, 0)

    for i in range(5):
        rover.send_speed()
        await trio.sleep(0.1)
        enc_counts_left.append(await rover.get_data(14))
        enc_counts_right.append(await rover.get_data(16))
        enc_intervals_left.append(await rover.get_data(28))
        enc_intervals_right.append(await rover.get_data(30))

    assert constant(enc_counts_left)
    assert constant(enc_counts_right)
    for i in enc_intervals_left:
        assert i == 0
    for i in enc_intervals_right:
        assert i == 0


def constant(L):
    return all(x == y for x, y in zip(L, L[1:]))


def strictly_increasing(L):
    return all(x < y for x, y in zip(L, L[1:]))


def strictly_decreasing(L):
    return all(x > y for x, y in zip(L, L[1:]))


async def test_encoder_intervals_forward(rover):
    enc_counts_left = []
    enc_counts_right = []
    enc_intervals_left = []
    enc_intervals_right = []

    rover.set_motor_speeds(0.1, 0.1, 0)

    for i in range(20):
        rover.send_speed()
        time.sleep(0.1)
        enc_counts_left.append(await rover.get_data(14))
        enc_counts_right.append(await rover.get_data(16))
        enc_intervals_left.append(await rover.get_data(28))
        enc_intervals_right.append(await rover.get_data(30))

    # note this test may fail if the rover wheels have any significant resistance, since this may cause the motors to "kick" backwards
    assert strictly_increasing(enc_counts_left)
    assert strictly_increasing(enc_counts_right)

    assert 0 < statistics.mean(enc_intervals_left) < 50000
    assert 0.001 < statistics.stdev(enc_intervals_left) < statistics.mean(enc_intervals_left) * 0.1


async def test_fan_speed(rover):
    for i in range(0, 241, 20):
        rover.set_fan_speed(i)
        await trio.sleep(0.1)
        assert i == await rover.get_data(48)


async def test_encoder_intervals_backward(rover):
    enc_counts_left = []
    enc_counts_right = []
    enc_intervals_left = []
    enc_intervals_right = []

    rover.set_motor_speeds(-0.1, -0.1, 0)

    for i in range(20):
        rover.send_speed()
        await trio.sleep(0.1)
        data = await rover.get_data_items([14, 16, 28, 30])
        enc_counts_left.append(data[14])
        enc_counts_right.append(data[16])
        enc_intervals_left.append(data[28])
        enc_intervals_right.append(data[30])

    # note this test may fail if the rover wheels have any significant resistance, since this may cause the motors to "kick" backwards
    assert strictly_decreasing(enc_counts_left)
    assert strictly_decreasing(enc_counts_right)

    assert 0 < statistics.mean(enc_intervals_left) < 50000
    assert 0.001 < statistics.stdev(enc_intervals_left) < statistics.mean(enc_intervals_left) * 0.1


async def test_currents(rover):
    battery_current_a = await rover.get_data(42)
    battery_current_b = await rover.get_data(44)
    battery_current_a_i2c = await rover.get_data(68)
    battery_current_b_i2c = await rover.get_data(70)

    # must agree within 5% or 200mA
    assert isclose(battery_current_a, abs(battery_current_a_i2c), rel_tol=0.05, abs_tol=0.2)
    assert isclose(battery_current_b, abs(battery_current_b_i2c), rel_tol=0.05, abs_tol=0.2)

    battery_current_total = await rover.get_data(0)
    assert isclose(battery_current_a + battery_current_b, battery_current_total, rel_tol=0.05, abs_tol=.05)


async def test_soc(rover):
    battery_soc_a = await rover.get_data(34)
    battery_soc_b = await rover.get_data(34)
    assert 0 <= battery_soc_a <= 1
    assert 0 <= battery_soc_b <= 1
    assert isclose(battery_soc_a, battery_soc_b, rel_tol=0.1, abs_tol=0.1)


async def test_currents_charging(rover):
    is_charging = await rover.get_data(38)
    assert is_charging in [True, False]

    battery_current_a_i2c = await rover.get_data(68)
    battery_current_b_i2c = await rover.get_data(70)
    if is_charging:
        # charging battery has positive current
        assert battery_current_a_i2c >= 0
        assert battery_current_b_i2c >= 0
    else:
        # discharging battery should have negative current
        assert battery_current_a_i2c <= 0
        assert battery_current_b_i2c <= 0


async def test_voltages(rover):
    battery_voltage_a = await rover.get_data(24)
    battery_voltage_a_i2c = await rover.get_data(64)
    assert 14 < battery_voltage_a_i2c < 18  # voltage should be between 14-18 V
    assert 14 < battery_voltage_a < 18
    # must agree within 5% or 50mV
    assert isclose(battery_voltage_a_i2c, battery_voltage_a, rel_tol=0.05, abs_tol=0.05)

    battery_voltage_b = await rover.get_data(26)
    battery_voltage_b_i2c = await rover.get_data(66)
    assert 14 < battery_voltage_b_i2c < 18
    assert 14 < battery_voltage_b < 18
    # must agree within 5% or 50mV
    assert isclose(battery_voltage_b_i2c, battery_voltage_b, rel_tol=0.05, abs_tol=0.05)


async def test_temperatures(rover):
    fan_temp = await rover.get_data(20)
    assert 10 < fan_temp < 55

    battery_temp_a = await rover.get_data(60)
    battery_temp_b = await rover.get_data(62)

    # Rated operating temperature check batteries are between 10 and 55 degrees C
    assert 10 < battery_temp_a < 55
    assert 10 < battery_temp_b < 55
    assert abs(battery_temp_a - battery_temp_b) < 5



from functools import partial
import logging
from typing import Callable, List
import warnings

import serial
import serial.tools
import serial.tools.list_ports
import trio


class OpenRoverWarning(RuntimeWarning):
    pass


class SerialTrio(trio.abc.AsyncResource):
    _serial: serial.Serial
    _inbound_high_water: int = 4000
    _outbound_high_water: int = 8000

    def __init__(self, port, **serial_kwargs):
        self.port = port
        self.serial_kwargs = dict(write_timeout=0, inter_byte_timeout=None, timeout=0, **serial_kwargs)
        try:
            self._serial = serial.Serial(self.port, **self.serial_kwargs)
        except serial.SerialException as e:
            if 'FileNotFoundError' in e.args[0]:
                raise RuntimeError("Could not connect to serial device - file not found. Is it connected?", self.port) from e
            if 'PermissionError' in e.args[0]:
                raise RuntimeError("Could not connect to serial device - permission error. Is it open in another process? Does this user have OS permission?", self.port) from e

    def _read_bytes_nowait(self, max):
        if self._inbound_high_water <= self._serial.in_waiting:
            warnings.warn(f'Incoming buffer is backlogged. Data may be lost. {self._serial.in_waiting} bytes')
        return self._serial.read(max)

    async def read_until(self, terminator):
        terminator = bytes(terminator)
        assert terminator != b''
        line = bytearray()
        try:
            while not line.endswith(terminator):
                line.extend(self._read_bytes_nowait(1))
                await trio.sleep(0.001)
            return bytes(line)
        except trio.Cancelled:
            logging.exception(f'abandoning data {line}')
            self._serial.cancel_read()
            raise

    async def read_exactly(self, count):
        line = bytearray()
        while len(line) < count:
            line.extend(self._read_bytes_nowait(count - len(line)))
            await trio.sleep(0.001)
        return bytes(line)

    async def write(self, data):
        self._serial.write(data)
        if self._outbound_high_water <= self._serial.out_waiting:
            warnings.warn(f'Outgoing buffer is backlogged. Data may be lost. {self._serial.out_waiting} bytes')
        try:
            await self.flush()
        except trio.Cancelled():
            self._serial.cancel_write()
            raise

    async def flush(self):
        while self._serial.out_waiting:
            await trio.sleep(0.001)

    async def aclose(self):
        try:
            while self._serial.out_waiting:
                await trio.sleep(0.001)
        finally:
            self._serial.close()
            assert not self._serial.is_open


def get_possible_rover_devices() -> List[Callable[[], SerialTrio]]:
    ports = []
    for comport in serial.tools.list_ports.comports():
        if comport.manufacturer != 'FTDI':
            continue
        ports.append(comport.device)
        DEFAULT_SERIAL_KWARGS = dict(baudrate=57600, stopbits=1)
    return [partial(SerialTrio, port, **DEFAULT_SERIAL_KWARGS) for port in ports]


def open_first_possible_rover_device():
    """use this as `async with open_first_possible_rover_device() as c:` """
    return next(iter(get_possible_rover_devices()))()



from concurrent.futures import Future
from typing import Any, Dict, Iterable, MutableMapping

from .openrover_protocol import OpenRoverProtocol
from .serial_trio import open_first_possible_rover_device
from .util import OpenRoverException


class OpenRover:
    _motor_left = 0
    _motor_right = 0
    _motor_flipper = 0
    _port = None
    _next_data = None
    _rover_protocol = None

    def __init__(self, port=None):
        """An OpenRover object """
        raise NotImplementedError

        self._motor_left = 0
        self._motor_right = 0
        self._motor_flipper = 0
        self._port = port
        self._next_data: MutableMapping[int, Future] = dict()

    async def aopen(self):
        port = self._port
        if port is None:
            port = await open_first_possible_rover_device()

        r, w = await self._connection.aopen()
        self._rover_protocol = OpenRoverProtocol(r, w)

    async def __aenter__(self):
        await self.aopen()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()

    async def process_data(self):
        try:
            async for k, v in self._rover_protocol.iter_packets():
                old_future = self._next_data.pop(k, None)
                if old_future is None:
                    raise RuntimeWarning('value was not expected %s: %s', k, v)
                else:
                    try:
                        old_future.set_result(v)
                    except InvalidStateError:
                        pass
        except Exception as e:
            raise

    async def consume_data(self, q):
        while True:
            key, value = await q.get()
            self._next_data[key].set_result(value)

    async def aclose(self):
        self._process_data_task.cancel()
        return await self._connection.aclose()

    def set_motor_speeds(self, left, right, flipper):
        assert -1 <= left <= 1
        assert -1 <= right <= 1
        assert -1 <= flipper <= 1
        self._motor_left = left
        self._motor_right = right
        self._motor_flipper = flipper

    async def send_command(self, arg1, arg2):
        await self._rover_protocol.write(self._motor_left, self._motor_right, self._motor_flipper, arg1, arg2)

    async def send_speed(self):
        await self.send_command(0, 0)

    async def set_fan_speed(self, fan_speed):
        await self.send_command(20, fan_speed)

    async def flipper_calibrate(self):
        await self.send_command(250, 250)

    def request_data(self, index):
        self.send_command(10, index)

    async def get_data(self, index):
        """Get the next value for the given data index."""
        old_future = self._next_data.get(index)
        if old_future is not None:
            old_future.cancel()

        new_future = asyncio.get_event_loop().create_future()
        self._next_data[index] = new_future
        await asyncio.wait([self.send_command(10, index), new_future], timeout=1)
        return await asyncio.wait_for(new_future, 1)

    async def get_data_items(self, indices: Iterable[int]) -> Dict[int, Any]:
        keys = list(set(indices))
        futures = []
        for i in keys:
            old_future = self._next_data.get(i)
            if old_future is not None:
                old_future.cancel()
            f = asyncio.get_event_loop().create_future()
            self._next_data[i] = f
            self.request_data(i)
            futures.append(f)
        values = await asyncio.wait_for(asyncio.gather(*futures), timeout=0.2)
        return dict(zip(keys, values))


async def get_openrover_version(port):
    async def get_openrover_version_untimed():
        async with OpenRoverConnection(port, open_timeout=3) as (reader, writer):
            protocol = OpenRoverProtocol(reader, writer)
            await protocol.write(0, 0, 0, 10, 40)
            async for k, v in protocol.iter_packets():
                if k == 40:
                    return v

    try:
        version = await get_openrover_version_untimed()
        return version
    except Exception as e:
        raise OpenRoverException(f'Did not respond to request for OpenRover version') from e



from .openrover_protocol import OpenRoverProtocol
from .serial_trio import open_first_possible_rover_device
from .util import OpenRoverException

name = 'openrover'

# rover.OpenRover is not yet ready for primetime
__all__ = ['OpenRoverException', 'OpenRoverProtocol', 'open_first_possible_rover_device']



from importlib.util import find_spec

import pytest


def main():
    openrover_search_paths = find_spec('openrover').submodule_search_locations
    return pytest.main(openrover_search_paths)


if __name__ == '__main__':
    exit(main())
